#ifndef DATALOGDATALOGTREE_H
#define DATALOGDATALOGTREE_H
#include <vector>
#include <iostream>
#include <map>
#include <string>

#include "langtree.h"
#include "dlognodevalue.h"
namespace datalog {

    /**
    DataLog tree is the language tree used by datalog parser to put language elements in.
    DLogTreeNode is the node in the DatalogTree;

    	@author xujian <shelbyxu@gmail.com>
    */
class DlogTreeNode;
typedef enum DlogTreeNodeType {
    N_ROOT = 0,
    N_RULE = 1,
    N_HEAD = 2,
    N_BODY = 3,
    N_SUBGOAL = 4,
    N_VAR = 5,
    N_NUMCONST = 6,
    N_STRCONST = 7,
    N_PRED = 8,
    N_RELATION = 9, //Relation |_ RELSYM, |_ VAR, |_ VAR, ...
    N_AGG = 10, // Aggregation
    N_RELSYM = 11, // Relation Symbol
    N_AGGFUNC = 12, // aggregation function name
    N_EXPR = 13,
    N_BOOLOP = 14,
    N_MATHOP  =15,
    N_PREDLIST = 16,
    N_UNDEFINED = 100
} DlogTreeNodeType;

typedef map<DlogTreeNodeType, string> TreeNodeTypeMap;
typedef vector<DlogTreeNode*> Children;
class DlogNodeValue;
class DlogTreeNode : public LangTreeNode {
  public:
        DlogTreeNode();
        DlogTreeNode(const DlogTreeNode& node);
        DlogTreeNode(const DlogTreeNodeType _type, DlogNodeValue v=DlogNodeValue());
        virtual ~DlogTreeNode();
        //Get and Set
        const DlogTreeNodeType getNodeType() const ;
        const DlogTreeNode& getChild(const int idx) const;
        const int getChildrenSize() const ;
        // return the non-null entries in a child
        const int getEffChildrenSize() const;
        DlogTreeNode* retChild(const int idx) const;
        const DlogTreeNode& getParent() const;
        DlogTreeNode* retParent() const;
        void setParent(const DlogTreeNode* parent);
        Children* retChildren() {
            return &children;
        };
        void setType (DlogTreeNodeType t);
        void setValue(DlogNodeValue v);
        void setValue(const string& v);
        const DlogNodeValue& getValue() const;
        bool isComplete() const {
            return complete;
        };
        bool equal(const DlogTreeNode& rhs) const;
        bool equal(const string& strval) const;
        int appendChild(DlogTreeNode* node);
        int pushChild(DlogTreeNode* node);
        /* detach a child, 3 modes:
          mode 0: remove from children and delete
          mode 1: remove from children but don't delete it
          mode 2: detach from children, delete and set that entry to NULL
          mode 3: detach from children, set entry to NULL but don't delete

          for mode 1 and 3, user should call retChildren to avoid memory leak.
        */
        int detachChild(const int idx, const int mode);
        // pack is to remove all NULL place holders in the children;
        void pack();
        void clear();
        int replaceSubtree(const int idx, DlogTreeNode* node);
        // Debugging and printing part.
        // PrintTreeNode, with indent
        void PrintTreeNode(ostream& out, int indent) const;
        //Print tree recursively
        void PrintSubTree(ostream& out, int indent) const;
        //mid-order travalsal for printing binary expression.
        void PrintSubTreeMid(string& dlog) const;

        DlogTreeNode* findNodeInSubTree(DlogTreeNodeType type);
        // find all nodes in the subtree(including itself) that is equal to s, results in res.
        void findAllNodesV(const DlogTreeNode* s,
                           vector<DlogTreeNode*>& res);
        void findAllNodesT(const DlogTreeNodeType t,
                           vector<DlogTreeNode*>& res);
        DlogTreeNode* deepCopy() const;
        int findChildbyPointer(DlogTreeNode* ptr);
        int findSelfIdx();
        //Serialize to disk
        void Dump(ostream& out) const;
        static DlogTreeNode* Read(istream& in, int& cs);
  private:
        static void InitNodeTypeMap();
        void Init() {
            type = N_UNDEFINED;
            children.clear();
            parent = NULL;
            InitNodeTypeMap();
        }
        DlogTreeNodeType type;
        Children children;
        DlogTreeNode* parent;
        DlogNodeValue value;
        static TreeNodeTypeMap nodetypemap;
        static bool inited;
        // Status of completeness, useful in parsing.
        bool complete;
};

class DatalogTree : public LangTree {
  public:
        DatalogTree();
        DatalogTree(const DatalogTree& t);
        ~DatalogTree();
        void Init();
        DlogTreeNode* retRoot();
        const DlogTreeNode* getRoot() const {
            return &root;
        }
        void clearTree();
        //Find a TreeNode using depth first search
        DlogTreeNode* findNode(DlogTreeNodeType type,
			       DlogTreeNode* start=NULL) const;
        int findAllNodesV(const DlogTreeNode *s,
                          vector<DlogTreeNode*>& res,
                          DlogTreeNode* start = NULL);
        int findAllNodesT(const DlogTreeNodeType t,
                          vector<DlogTreeNode*>& res,
                          DlogTreeNode* start = NULL);
        int findRelbyName(const string& name,
                          vector<int>& res,
                          DlogTreeNode* start = NULL) const;
        //Return the number of occurances.
        int substitude(const DlogTreeNodeType type, 
		       const string& value,
                       DlogTreeNode* target,
                       DlogTreeNode* start=NULL);
        // Recover from tree to datalog.
        void printDatalog() const ;
        void printDatalog(ostream& out) const;
        void printTree() const;

        // return semantic sub-structures of a datlog rule.
        DlogTreeNode* getHead() const ;
        DlogTreeNode* getBody() const ;
				DlogTreeNode* getSubG() const ;

        // dump to and read from disk
        void Dump(ostream& out) const ;
        static DatalogTree* Read(istream& in);
				// deep copy.
				DatalogTree* DeepCopy() const;
  private:
        DlogTreeNode root;
};

}

#endif
